﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Drawing.Drawing2D;

namespace SplitTile {
    public partial class Form1 : Form {
        public string dir;
        public class page {
            public string filename;
            public Bitmap bmp;
            public int divw, divh;
            public PointF top, bottom;
            public int rot;
            public int sort;
        };
        public List<page> pages = new List<page>();
        bool bycode = false;

        public Form1() {
            InitializeComponent();
        }

        void LoadDirectory(string Dir) {
            if (Dir == "") return;
            //ディレクトリから画像の読み込み
            this.dir = Dir;
            List<string> files = new List<string>();
            foreach (string pat in new string[] { "*.jpg", "*.jpeg", "*.png", "*.gif", "*.bmp" }) {
                foreach (string f in Directory.GetFiles(Dir, pat)) files.Add(f);
            }
            files.Sort();
            pages.Clear();
            foreach (string f in files) {
                page p = new page();
                p.filename = Path.GetFileName(f);
                p.bmp = new Bitmap(f);
                p.divh = p.divw = 5;
                pages.Add(p);
            }
            listBox1.Items.Clear();
            listBox1.Items.AddRange(files.ToArray());

        }

        private void Form1_Load(object sender, EventArgs e) {
            if (folderBrowserDialog1.ShowDialog() != DialogResult.OK)
                Application.Exit();
            LoadDirectory(folderBrowserDialog1.SelectedPath);
            if (listBox1.Items.Count == 0) {
                MessageBox.Show("画像ファイルが見つからないため終了します。");
                Application.Exit();
            }
            toolStripStatusLabel1.Text = "画像エリアを右クリック後、Shiftを押しながらカーソルキーで位置を微調整、Ctrlを押しながらカーソルキーでサイズを微調整できます";

        }

        private RectangleF RotZoom(page p) {
            //画像の回転
            float s = (float)Math.Sin(p.rot * Math.PI / 1800);
            float c = (float)Math.Cos(p.rot * Math.PI / 1800);

            PointF[] ps = new PointF[4];
            ps[0] = new PointF(0, 0);
            ps[1] = new PointF(p.bmp.Width * c, p.bmp.Width * s);
            ps[2] = new PointF(p.bmp.Width * c - p.bmp.Height * s, p.bmp.Width * s + p.bmp.Height * c);
            ps[3] = new PointF(-p.bmp.Height * s, p.bmp.Height * c);
            float le, ri, to, bo;
            le = ri = ps[0].X;
            to = bo = ps[0].Y;
            for (int i = 1; i <= 3; i++) {
                le = Math.Min(le, ps[i].X);
                ri = Math.Max(ri, ps[i].X);
                to = Math.Min(to, ps[i].Y);
                bo = Math.Max(bo, ps[i].Y);
            }
            return new RectangleF(le, to, ri - le, bo - to);
        }

        private void pictureBox1_Paint(object sender, PaintEventArgs e) {
            Graphics g = e.Graphics;
            g.Clear(Color.Gray);
            if (listBox1.SelectedIndex == -1) return;
            page p = pages[listBox1.SelectedIndex];
            RectangleF rect = RotZoom(p);

            float ws = pictureBox1.ClientRectangle.Width / rect.Width;
            float hs = pictureBox1.ClientRectangle.Height / rect.Height;
            float scale = Math.Min(ws, hs);

            g.ResetTransform();
            g.TranslateTransform(-rect.Left, -rect.Top);
            g.ScaleTransform(scale, scale, MatrixOrder.Append);
            g.RotateTransform(p.rot / 10.0f);
            g.DrawImage(p.bmp, 0, 0, p.bmp.Width, p.bmp.Height);


            if (p.top.X != p.bottom.X && p.top.Y != p.bottom.Y) {
                rect = new RectangleF();
                rect.X = Math.Min(p.top.X, p.bottom.X);
                rect.Y = Math.Min(p.top.Y, p.bottom.Y);
                rect.Width = Math.Max(p.top.X, p.bottom.X) - rect.X;
                rect.Height = Math.Max(p.top.Y, p.bottom.Y) - rect.Y;

                //枠線
                g.ResetTransform();
                Pen bp = new Pen(Color.Black, 3);
                bp.DashStyle = DashStyle.Dot;
                Pen wp = new Pen(Color.White, 3);
                g.DrawRectangle(wp, rect.X * scale, rect.Y * scale, rect.Width * scale, rect.Height * scale);
                g.DrawRectangle(bp, rect.X * scale, rect.Y * scale, rect.Width * scale, rect.Height * scale);
                // 中の分割線
                wp.Width = bp.Width = 2;
                for (int x = 1; x < p.divw; x++) {
                    g.DrawLine(wp, (rect.X + rect.Width * x / p.divw) * scale, rect.Y * scale,
                        (rect.X + rect.Width * x / p.divw) * scale, rect.Bottom * scale);
                    g.DrawLine(bp, (rect.X + rect.Width * x / p.divw) * scale, rect.Y * scale,
                        (rect.X + rect.Width * x / p.divw) * scale, rect.Bottom * scale);
                }
                for (int y = 1; y < p.divh; y++) {
                    g.DrawLine(wp, rect.X * scale, (rect.Y + rect.Height * y / p.divh) * scale,
                        rect.Right * scale, (rect.Y + rect.Height * y / p.divh) * scale);
                    g.DrawLine(bp, rect.X * scale, (rect.Y + rect.Height * y / p.divh) * scale,
                        rect.Right * scale, (rect.Y + rect.Height * y / p.divh) * scale);
                }
                toolStripStatusLabel2.Text = String.Format("({0},{1})-({2},{3})", (int)p.top.X, (int)p.top.Y, (int)p.bottom.X, (int)p.bottom.Y);
            }
            else {
                toolStripStatusLabel2.Text = "未設定";
            }


        }

        private void listBox1_SelectedIndexChanged(object sender, EventArgs e) {
            if (listBox1.SelectedIndex == -1) return;
            page p = pages[listBox1.SelectedIndex];
            bycode = true;
            numWidth.Value = p.divw;
            numHeight.Value = p.divh;
            numRot.Value = p.rot;
            if (p.sort == 0) radioZsort.Checked = true;
            if (p.sort == 1) radioNsort.Checked = true;
            if (p.sort == 2) radioRevNsort.Checked = true;

            bycode = false;

            pictureBox1.Refresh();


        }

        private void pictureBox1_ClientSizeChanged(object sender, EventArgs e) {
            pictureBox1.Refresh();
        }

        private void numWidth_ValueChanged(object sender, EventArgs e) {
            //GUIでパラメータ調整
            if (listBox1.SelectedIndex == -1) return;
            page p = pages[listBox1.SelectedIndex];

            if (!bycode) {
                p.divw = (int)numWidth.Value;
                p.divh = (int)numHeight.Value;
                p.rot = (int)numRot.Value;
                if (radioZsort.Checked) p.sort = 0;
                if (radioNsort.Checked) p.sort = 1;
                if (radioRevNsort.Checked) p.sort = 2;
                pictureBox1.Refresh();
            }

        }

        private void butClear_Click(object sender, EventArgs e) {
            if (listBox1.SelectedIndex == -1) return;
            page p = pages[listBox1.SelectedIndex];
            p.top.X = p.top.Y = 0;
            p.bottom.X = p.bottom.Y = 0;
            string fn = listBox1.SelectedItem.ToString();
            if (fn.StartsWith("*")) fn = fn.Substring(1);
            listBox1.Items[listBox1.SelectedIndex] = fn;
            pictureBox1.Refresh();

        }


        private void pictureBox1_MouseDown(object sender, MouseEventArgs e) {
            pictureBox1.Focus();
            if (e.Button != MouseButtons.Left) return;
            //分割位置の左上点指定

            if (listBox1.SelectedIndex == -1) return;
            page p = pages[listBox1.SelectedIndex];
            RectangleF rect = RotZoom(p);
            float ws = pictureBox1.ClientRectangle.Width / rect.Width;
            float hs = pictureBox1.ClientRectangle.Height / rect.Height;
            float scale = Math.Min(ws, hs);

            p.top.X = e.X / scale;
            p.top.Y = e.Y / scale;
            p.bottom = p.top;
            pictureBox1.Refresh();
        }

        private void pictureBox1_MouseMove(object sender, MouseEventArgs e) {
            if (e.Button != MouseButtons.Left) return;
            //分割位置の右下点指定
            if (listBox1.SelectedIndex == -1) return;
            page p = pages[listBox1.SelectedIndex];
            RectangleF rect = RotZoom(p);
            float ws = pictureBox1.ClientRectangle.Width / rect.Width;
            float hs = pictureBox1.ClientRectangle.Height / rect.Height;
            float scale = Math.Min(ws, hs);

            p.bottom.X = e.X / scale;
            p.bottom.Y = e.Y / scale;
            pictureBox1.Refresh();

        }

        private void CopyPage(int from, int to) {
            string fn = listBox1.Items[from].ToString();
            if (fn.StartsWith("*")) {
                page p = pages[to];
                page op = pages[from];
                p.top = op.top;
                p.bottom = op.bottom;
                p.divh = op.divh;
                p.divw = op.divw;
                p.rot = op.rot;

                pictureBox1.Refresh();

                fn = listBox1.Items[to].ToString();
                if (fn.StartsWith("*")) fn = fn.Substring(1);
                listBox1.Items[to] = "*"+fn;
            }
            else {
                toolStripStatusLabel1.Text = "未設定の画像からのコピーはできません。";
            }
        }

        private void butCopy1_Click(object sender, EventArgs e) {
            if (listBox1.SelectedIndex >= 1) CopyPage(listBox1.SelectedIndex-1, listBox1.SelectedIndex);

        }

        private void butCopy2_Click(object sender, EventArgs e) {
            if (listBox1.SelectedIndex >= 2) CopyPage(listBox1.SelectedIndex-2, listBox1.SelectedIndex);
        }

        private void btnCopyAll_Click(object sender, EventArgs e){
            if (listBox1.SelectedIndex >= 0)
            {
                int i;
                for (i = 0; i < listBox1.Items.Count; i++) {
                    toolStripStatusLabel1.Text = String.Format("処理中：{0}/{1}", i + 1, listBox1.Items.Count);
                    Application.DoEvents();
                    if (i != listBox1.SelectedIndex)
                        CopyPage(listBox1.SelectedIndex, i);
                }
            }

        }

        private void pictureBox1_MouseUp(object sender, MouseEventArgs e)
        {
            if (listBox1.SelectedIndex == -1) return;
            page p = pages[listBox1.SelectedIndex];
            // topがbottomの左上にくるようにする
            if (p.top.X > p.bottom.X) {
                float x = p.top.X;
                p.top.X = p.bottom.X; p.bottom.X = x;
            }
            if (p.top.Y > p.bottom.Y) {
                float y = p.top.Y;
                p.top.Y = p.bottom.Y; p.bottom.Y = y;
            }
            string fn = listBox1.SelectedItem.ToString();
            if (!fn.StartsWith("*"))
                listBox1.Items[listBox1.SelectedIndex] = "*" + fn;

        }

        private void Form1_KeyDown(object sender, KeyEventArgs e) {
            if (listBox1.SelectedIndex == -1) return;
            //キー操作による移動
            if (!e.Shift && !e.Control) {

                switch (e.KeyData & Keys.KeyCode) {
                    case Keys.C:
                        butClear_Click(null, null);
                        break;
                    case Keys.NumPad1:
                    case Keys.D1:
                        butCopy1_Click(null, null);
                        break;
                    case Keys.NumPad2:
                    case Keys.D2:
                        butCopy2_Click(null, null);
                        break;
                    case Keys.PageUp:
                        if (listBox1.SelectedIndex >= 1) listBox1.SelectedIndex--;
                        break;
                    case Keys.PageDown:
                        if (listBox1.SelectedIndex < listBox1.Items.Count-1) listBox1.SelectedIndex++;
                        break;
                    default:
                        return;
                }
                e.Handled = true;
                return;
            }
            e.Handled = true;
            page p = pages[listBox1.SelectedIndex];
            if (p.top == p.bottom) return;

            switch (e.KeyData & Keys.KeyCode) {
                case Keys.Left:
                    if (e.Shift) { p.top.X -= 1; p.bottom.X -= 1; }
                    else if (e.Control && p.bottom.X > p.top.X + 1) { p.bottom.X -= 1; }
                    break;
                case Keys.Right:
                    if (e.Shift) { p.top.X += 1; p.bottom.X += 1; }
                    else if (e.Control) { p.bottom.X += 1; }
                    break;
                case Keys.Up:
                    if (e.Shift) { p.top.Y -= 1; p.bottom.Y -= 1; }
                    else if (e.Control && p.bottom.Y > p.top.Y + 1) { p.bottom.Y -= 1; }
                    break;
                case Keys.Down:
                    if (e.Shift) { p.top.Y += 1; p.bottom.Y += 1; }
                    else if (e.Control) { p.bottom.Y += 1; }
                    break;
                default:
                    return;
            }

            pictureBox1.Refresh();

        }

        private void SplitPage(string tardir, page p) {
            RectangleF rect = RotZoom(p);
            string basename = Path.GetFileNameWithoutExtension(p.filename);

            Bitmap bmp = new Bitmap((int)rect.Width, (int)rect.Height, System.Drawing.Imaging.PixelFormat.Format32bppArgb);
            Graphics g = Graphics.FromImage(bmp);
            g.Clear(Color.Gray);

            int tw = (int)((p.bottom.X - p.top.X + 0.5) / p.divw);
            int th = (int)((p.bottom.Y - p.top.Y + 0.5) / p.divh);
            Bitmap spbmp = new Bitmap(tw, th, System.Drawing.Imaging.PixelFormat.Format32bppArgb);
            Graphics sg = Graphics.FromImage(spbmp);

            //まず回転画像を描画
            g.ResetTransform();
            g.TranslateTransform(-rect.Left, -rect.Top);
            g.RotateTransform(p.rot / 10.0f);
            g.DrawImage(p.bmp, 0, 0, p.bmp.Width, p.bmp.Height);

            for (int curp = 0; curp < p.divw * p.divh; curp++) {
                int tx = 0, ty = 0;
                if (p.sort == 0) { ty = curp / p.divw; tx = curp % p.divw; } //Z順
                else if (p.sort == 1) { ty = curp % p.divh; tx = (p.divh - 1) - curp % p.divh; } //N順
                else if (p.sort == 2) { ty = curp % p.divh; tx = curp % p.divh; } //逆N順

                string fn = String.Format("{0}\\{1}_{2:00}.png", tardir, basename, curp + 1);
                sg.DrawImageUnscaled(bmp, -(int)(p.top.X + tw * tx), -(int)(p.top.Y + th * ty));
                spbmp.Save(fn);
            }

        }

        private void DoSplit() {

            string tardir = dir + "\\split";

            // ディレクトリ作成
            try {
                if (!Directory.Exists(tardir)) {
                    Directory.CreateDirectory(tardir);
                }
            }
            catch (Exception e) {
                toolStripStatusLabel1.Text = e.Message;
                return;
            }

            for (int i = 0; i < listBox1.Items.Count; i++) {
                toolStripStatusLabel1.Text = String.Format("処理中：{0}/{1}", i + 1, listBox1.Items.Count);
                Application.DoEvents();
                if (pages[i].top == pages[i].bottom) continue;

                SplitPage(tardir, pages[i]);
            }
            toolStripStatusLabel1.Text = "完了！";

        }


        private void butSplit_Click(object sender, EventArgs e) {
            DoSplit();
        }



    }
}
